# Copyright (C) 2008-2025 Luis Falcon <falcon@gnuhealth.org>
# Copyright (C) 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from sql.aggregate import Count
from sql.functions import DateTrunc
from datetime import date, datetime
from trytond.report import Report
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.modules.health.core import parse_compute_age
from dateutil.relativedelta import relativedelta

__all__ = ['InstitutionSummaryReport']


class InstitutionSummaryReport(Report):
    __name__ = 'gnuhealth.summary.report'

    @classmethod
    def get_population_with_no_dob(cls):
        """ Return Total Number of living people in the system
        without a date of birth"""
        pool = Pool()
        Party = pool.get('party.party')

        return Party.search([
            ('is_person', '=', True),
            ('deceased', '!=', True),
            ('dob', '=', None),
        ], count=True)

    @classmethod
    def get_population(cls, date1, date2, gender, total):
        """ Return Total Number of living people in the system
        segmented by age group and gender"""
        pool = Pool()
        Party = pool.get('party.party')

        domain = [
            ('deceased', '!=', True),
            ('gender', '=', gender),
        ]

        if not total:
            domain.append(('dob', '>=', date2))
            domain.append(('dob', '<=', date1))

        return Party.search(domain, count=True)

    @classmethod
    def get_new_people(cls, start_date, end_date, in_health_system):
        """ Return Total Number of new registered persons alive """
        pool = Pool()
        Party = pool.get('party.party')

        domain = [
            ('activation_date', '>=', start_date),
            ('activation_date', '<=', end_date),
            ('deceased', '!=', True),
            ('is_person', '=', True),
        ]

        if in_health_system:
            domain.append(('is_patient', '=', True))

        return Party.search(domain, count=True)

    @classmethod
    def get_new_births(cls, start_date, end_date):
        """ Return birth certificates within that period """
        pool = Pool()
        BirthCertificate = pool.get('gnuhealth.birth_certificate')

        return BirthCertificate.search([
            ('dob', '>=', start_date),
            ('dob', '<=', end_date),
        ], count=True)

    @classmethod
    def get_new_deaths(cls, start_date, end_date):
        """ Return death certificates within that period """
        """ Truncate the timestamp of DoD to match a whole day"""
        pool = Pool()
        DeathCertificate = pool.get('gnuhealth.death_certificate')
        table = DeathCertificate.__table__()

        dod = DateTrunc('day', table.dod)

        cursor = Transaction().connection.cursor()
        cursor.execute(*table.select(
            Count(table.dod),
            where=((dod >= start_date) & (dod <= end_date))))
        return cursor.fetchone()

    @classmethod
    def get_evaluations(cls, start_date, end_date, dx, count=False):
        """ Return evaluation info """
        pool = Pool()

        Evaluation = pool.get('gnuhealth.patient.evaluation')
        start_date = datetime.strptime(str(start_date), '%Y-%m-%d')
        end_date = datetime.strptime(str(end_date), '%Y-%m-%d')
        end_date += relativedelta(hours=+23, minutes=+59, seconds=+59)

        clause = [
            ('evaluation_start', '>=', start_date),
            ('evaluation_start', '<=', end_date),
        ]

        if dx:
            clause.append(('diagnosis', '=', dx))

        return Evaluation.search(clause, count=count)

    @classmethod
    def count_evaluations(cls, start_date, end_date, dx):
        """ count diagnoses by groups """
        return cls.get_evaluation(start_date, end_date, dx, count=True)

    @classmethod
    def get_context(cls, records, header, data):
        context = super(
            InstitutionSummaryReport, cls).get_context(records, header, data)

        start_date = data['start_date']
        end_date = data['end_date']

        context['demographics'] = data['demographics']
        context['patient_evaluations'] = data['patient_evaluations']

        context['start_date'] = data['start_date']
        context['end_date'] = data['end_date']

        # Demographics
        today = date.today()

        context[''.join(['p', 'total_', 'f'])] = \
            cls.get_population(None, None, 'f', total=True)

        context[''.join(['p', 'total_', 'm'])] = \
            cls.get_population(None, None, 'm', total=True)

        # Living people with NO date of birth
        context['no_dob'] = \
            cls.get_population_with_no_dob()

        # Build the Population Pyramid for registered people
        for age_group in range(0, 21):
            date1 = today - relativedelta(years=(age_group * 5))
            date2 = today - relativedelta(years=((age_group * 5) + 5), days=-1)

            context[''.join(['p', str(age_group), 'f'])] = \
                cls.get_population(date1, date2, 'f', total=False)
            context[''.join(['p', str(age_group), 'm'])] = \
                cls.get_population(date1, date2, 'm', total=False)

        # Count those lucky over 105 years old :)
        date1 = today - relativedelta(years=105)
        date2 = today - relativedelta(years=200)

        context['over105f'] = \
            cls.get_population(date1, date2, 'f', total=False)
        context['over105m'] = \
            cls.get_population(date1, date2, 'm', total=False)

        # Count registered people, and those within the system of health
        context['new_people'] = \
            cls.get_new_people(start_date, end_date, False)
        context['new_in_health_system'] = \
            cls.get_new_people(start_date, end_date, in_health_system=True)

        # New births
        context['new_births'] = \
            cls.get_new_births(start_date, end_date)

        # New deaths
        context['new_deaths'] = \
            cls.get_new_deaths(start_date, end_date)

        # Get evaluations within the specified date range

        context['evaluations'] = \
            cls.get_evaluations(start_date, end_date, None)

        evaluations = cls.get_evaluations(start_date, end_date, None)

        eval_dx = []
        non_dx_eval = 0
        non_age_eval = 0
        eval_f = 0
        eval_m = 0

        # Global Evaluation info
        for evaluation in evaluations:
            if evaluation.diagnosis:
                eval_dx.append(evaluation.diagnosis)
            else:
                # Increase the evaluations without Dx counter
                non_dx_eval += 1
            if (evaluation.gender == 'f'):
                eval_f += 1
            else:
                eval_m += 1

            if not evaluation.computed_age:
                non_age_eval += 1

        context['non_dx_eval'] = non_dx_eval
        context['eval_num'] = len(evaluations)
        context['eval_f'] = eval_f
        context['eval_m'] = eval_m
        context['non_age_eval'] = non_age_eval

        # Create a set to work with single diagnoses
        # removing duplicate entries from eval_dx

        unique_dx = set(eval_dx)

        summary_dx = []
        # Traverse the evaluations with Dx to get the key values
        for dx in unique_dx:
            # unique_evaluations holds the evaluations with
            # the same diagnostic
            unique_evaluations = cls.get_evaluations(start_date, end_date, dx)

            group_1 = group_2 = group_3 = group_4 = group_5 = 0
            group_1f = group_2f = group_3f = group_4f = group_5f = 0

            total_evals = len(unique_evaluations)

            new_conditions = 0

            for unique_eval in unique_evaluations:

                if (unique_eval.computed_age):

                    # Strip to get the raw year
                    age_year = parse_compute_age(unique_eval.computed_age)[0]

                    # Age groups in this diagnostic
                    if (age_year < 5):
                        group_1 += 1
                        if (unique_eval.gender == 'f'):
                            group_1f += 1
                    if (age_year in range(5, 14)):
                        group_2 += 1
                        if (unique_eval.gender == 'f'):
                            group_2f += 1
                    if (age_year in range(15, 45)):
                        group_3 += 1
                        if (unique_eval.gender == 'f'):
                            group_3f += 1
                    if (age_year in range(46, 60)):
                        group_4 += 1
                        if (unique_eval.gender == 'f'):
                            group_4f += 1
                    if (age_year > 60):
                        group_5 += 1
                        if (unique_eval.gender == 'f'):
                            group_5f += 1

                # Check for new conditions vs followup / chronic checkups
                # in the evaluation with a particular diagnosis

                if (unique_eval.visit_type == 'new'):
                    new_conditions += 1

            eval_dx = {
                'diagnosis': unique_eval.diagnosis.rec_name,
                'age_group_1': group_1, 'age_group_1f': group_1f,
                'age_group_2': group_2, 'age_group_2f': group_2f,
                'age_group_3': group_3, 'age_group_3f': group_3f,
                'age_group_4': group_4, 'age_group_4f': group_4f,
                'age_group_5': group_5, 'age_group_5f': group_5f,
                'total': total_evals, 'new_conditions': new_conditions,
            }

            # Append into the report list the resulting
            # dictionary entry

            summary_dx.append(eval_dx)

        context['summary_dx'] = summary_dx

        return context



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

try:
    from trytond.modules.health_dentistry.tests.test_health_dentistry \
        import suite
except ImportError:
    from .test_health_dentistry import suite

__all__ = ['suite']



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                        HEALTH IMAGING package                         #
#              __init__.py: Package declaration file                    #
#########################################################################

from trytond.pool import Pool
from . import health_imaging
from . import wizard
from . import sequences


def register():
    Pool.register(
        health_imaging.ImagingTestType,
        health_imaging.ImagingTest,
        health_imaging.ImagingTestRequest,
        health_imaging.ImagingTestResult,
        wizard.wizard_health_imaging.RequestImagingTest,
        wizard.wizard_health_imaging.RequestPatientImagingTestStart,
        sequences.GnuHealthSequences,
        sequences.ImagingRequestSequence,
        sequences.ImagingTestSequence,
        module='health_imaging', type_='model')

    Pool.register(
        wizard.wizard_health_imaging.WizardGenerateResult,
        wizard.wizard_health_imaging.RequestPatientImagingTest,
        module='health_imaging', type_='wizard')



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                 HEALTH PEDIATRICS_GROWTH_CHARTS_WHO package           #
#             health_pediatrics_growth_charts_who.py:main module        #
#########################################################################
from trytond.model import ModelView, ModelSQL, fields

__all__ = ['PediatricsGrowthChartsWHO']


class PediatricsGrowthChartsWHO(ModelSQL, ModelView):
    'Pediatrics Growth Chart WHO'
    __name__ = 'gnuhealth.pediatrics.growth.charts.who'

    indicator = fields.Selection([
        ('l/h-f-a', 'Length/height for age'),
        ('w-f-a', 'Weight for age'),
        ('bmi-f-a', 'Body mass index for age (BMI for age)'),
    ], 'Indicator', sort=False, required=True)
    indicator_str = indicator.translated('indicator')

    measure = fields.Selection([
        ('p', 'percentiles'),
        ('z', 'z-scores'),
    ], 'Measure')
    measure_str = measure.translated('measure')

    sex = fields.Selection([
        ('m', 'Male'),
        ('f', 'Female'),
    ], 'Sex')
    sex_str = sex.translated('sex')

    month = fields.Integer('Month')
    type = fields.Char('Type')
    value = fields.Float('Value')



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH SERVICES_IMAGING PACKAGE                   #
#                 Health_services_imaging.py: Main module               #
#########################################################################

from trytond.model import ModelView, fields
from trytond.pyson import Eval, Equal
from trytond.pool import Pool, PoolMeta
from trytond.i18n import gettext
from .exceptions import (NoServiceAssociated,
                         ServiceHasBeenUpdated)


__all__ = ['ImagingTestRequest']


""" Add Imaging order charges to service model """


class ImagingTestRequest(metaclass=PoolMeta):
    __name__ = 'gnuhealth.imaging.test.request'

    service = fields.Many2One(
        'gnuhealth.health_service', 'Service',
        domain=[('patient', '=', Eval('patient'))], depends=['patient'],
        states={'readonly': Equal(Eval('state'), 'done')},
        help="Service document associated to this Imaging Request")

    service_updated = fields.Selection((
        ('yes', 'Yes'),
        ('no', 'No'),
        ('unknown', 'Unknown'),
    ), 'Service updated', sort=False)

    @classmethod
    def default_service_updated(self):
        return 'unknown'

    @classmethod
    def __setup__(cls):
        super(ImagingTestRequest, cls).__setup__()
        cls._buttons.update({
            'update_service': {
                'readonly': Equal(Eval('state'), 'done'),
            },
        })

    @classmethod
    @ModelView.button
    def update_service(cls, imaging_orders):
        imaging_order = imaging_orders[0]

        if not imaging_order.service:
            raise NoServiceAssociated(
                gettext('health_services_imaging.'
                        'msg_no_service_associated'))

        if imaging_order.service_updated == 'yes':
            raise ServiceHasBeenUpdated(
                gettext('health_services_imaging.'
                        'msg_service_has_been_updated'))

        pool = Pool()
        HealthService = pool.get('gnuhealth.health_service')

        hservice = []
        service_data = {}
        service_lines = []

        # Add the imaging order to the service document

        service_lines.append(('create', [{
            'product': imaging_order.requested_test.product.id,
            'desc': imaging_order.requested_test.product.rec_name,
            'qty': 1
        }]))

        hservice.append(imaging_order.service)

        description = "Services and Imaging"

        service_data['desc'] = description
        service_data['service_line'] = service_lines

        HealthService.write(hservice, service_data)

        cls.write(imaging_orders, {'service_updated': 'yes'})



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH FEDERATION package                         #
#                __init__.py: Package declaration file                  #
#########################################################################
"""
health_federation package: HMIS Federation manager

This package is responsible to communicate the GNU Health HMIS
with Thalamus, the Federation message server.

The main functionalities of the package are :
    * HMIS configuration to connect to Thalamus message server
    * Define the models that will participate on the Federation
    * Define which fields of each model will be shared on the Federation
    * Queue manager: Manages the local node queue to be sent to the Federation.
    * Defines the classes for each model and specific methods to interact
      with Thalamus
"""

from trytond.pool import Pool
from . import health_federation


def register():
    Pool.register(
        health_federation.FederationNodeConfig,
        health_federation.FederationQueue,
        health_federation.FederationObject,
        health_federation.PartyFed,
        health_federation.PoLFed,
        module='health_federation', type_='model')



# SPDX-FileCopyrightText: 2009-2013 Bertrand Chenal
# SPDX-FileCopyrightText: 2009-2016 B2CK
# SPDX-FileCopyrightText: 2009-2016 Cédric Krier
# SPDX-FileCopyrightText: 2009-2016 Tryton Foundation <info@tryton.org>
# SPDX-FileCopyrightText: 2016-2024 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2016-2024 Luis Falcón <falcon@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from trytond.model import fields
from trytond.pool import PoolMeta
from trytond.pyson import Bool, Eval

__all__ = ['User']


class User(metaclass=PoolMeta):
    __name__ = 'res.user'

    calendars = fields.One2Many('calendar.calendar', 'owner', 'Calendars')

    @classmethod
    def __setup__(cls):
        super(User, cls).__setup__()
        required = Bool(Eval('calendars'))
        if not cls.email.states.get('required'):
            cls.email.states['required'] = required
        else:
            cls.email.states['required'] = (
                cls.email.states['required'] | required)
        """ # Migration to GH 5.0. 'set' object has no attribute 'append
        if 'calendars' not in cls.email.depends:
            cls.email.depends.append('calendars')
        """



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#              wizard_check_immunization_status.py: wizard              #
#########################################################################
from trytond.wizard import Wizard, StateView, Button, StateAction
from trytond.model import ModelView, fields
from trytond.transaction import Transaction

__all__ = ['CheckImmunizationStatusInit', 'CheckImmunizationStatus']


class CheckImmunizationStatusInit(ModelView):
    'Check Immunization Status Init'
    __name__ = 'gnuhealth.check_immunization_status.init'
    immunization_schedule = fields.Many2One(
        "gnuhealth.immunization_schedule", "Schedule",
        required=True)


class CheckImmunizationStatus(Wizard):
    'Check Immunization Status'
    __name__ = 'gnuhealth.check_immunization_status'

    start = StateView(
        'gnuhealth.check_immunization_status.init',
        'health.view_check_immunization_status', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Immunization Status', 'check_immunization_status',
                   'tryton-ok', True),
        ])
    check_immunization_status = StateAction(
        'health.report_immunization_status')

    def do_check_immunization_status(self, action):
        return action, self.get_info()

    def get_info(self):

        return {
            'patient_id': Transaction().context.get('active_id'),
            'immunization_schedule_id': self.start.immunization_schedule.id
        }

    def transition_check_immunization_status(self):
        return 'end'



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2023-2025 Feng Shu (tumashu)
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#   health_report.py: Disease, Medication and Vaccination reports       #
#########################################################################
from datetime import datetime
from trytond.i18n import gettext

import io

from trytond.modules.health.core import get_institution_timezone

try:
    from PIL import Image
except ImportError:
    Image = None


__all__ = ['ReportDateAndTimeMixin',
           'ReportGettextMixin',
           'ReportImageToolMixin']


class ReportDateAndTimeMixin():

    @classmethod
    def get_context(cls, records, header, data):
        context = super(
            ReportDateAndTimeMixin, cls).get_context(
                records, header, data)

        context['print_datetime'] = datetime.now()
        context['tz'] = get_institution_timezone()

        return context


class ReportGettextMixin:
    '''Mixin gettext function and some strings need translation and
    used frequently.

    '''
    __slots__ = ()

    @classmethod
    def get_context(cls, records, header, data):
        context = super(ReportGettextMixin, cls).get_context(
            records, header, data)

        context['yes_str'] = gettext('health.msg_yes_str')
        context['no_str'] = gettext('health.msg_no_str')
        context['gettext'] = gettext

        return context


class ReportImageToolMixin:
    'Mixin to operate image in report.'
    __slots__ = ()

    @classmethod
    def image_keep_size(cls, image):
        """Return a tuple like: (image, mimetype, width, height), which is
        used in relatorio open document's image template, this tuple let
        image keep its pt size.
        """
        try:
            img = Image.open(io.BytesIO(image))
            width, height = img.size

            w = str(width) + 'pt'
            h = str(height) + 'pt'
            mimetype = None

        except BaseException:

            w = None
            h = None
            mimetype = None

        return (image, mimetype, w, h)

    @classmethod
    def image_resize(cls, image, max_width, max_height, unit='cm'):
        """Return a tuple like: (image, mimetype, width, height), which is
        used in relatorio open document's image template, this tuple let
        image showed in odt file keep ratio, and width < max_width, height
        < max_height.
        """
        try:
            img = Image.open(io.BytesIO(image))
            orig_width, orig_height = img.size
            orig_ratio = float(orig_height / orig_width)
            ratio = max_height / max_width

            if orig_ratio >= ratio:
                width = max_height / orig_ratio
                height = max_height
            else:
                width = max_width
                height = max_width * orig_ratio

            w = str(width) + unit
            h = str(height) + unit
            mimetype = None

        except BaseException:

            w = str(max_width) + unit
            h = str(max_height) + unit
            mimetype = None

        return (image, mimetype, w, h)

    @classmethod
    def image_crop(cls, image, width, height, unit='cm'):
        """Center-crop image and return a tuple like:

            (new_image, mimetype, width, height)

        which is used in relatorio open document's image template, this
        tuple let image showed in odt file keep ratio and size = (width,
        height).

        """
        try:
            ratio = height / width
            image_info = cls._image_crop_to_ratio(image, ratio)
            image = image_info.get('image')
            mimetype = image_info.get('mimetype')
        except BaseException:
            image = image
            mimetype = None

        w = str(width) + unit
        h = str(height) + unit
        return (image, mimetype, w, h)

    @classmethod
    def _image_crop_to_ratio(cls, image, ratio):
        """ Center-crop an image, make it conform to the ratio,
        This function is useful to adjust ID card photo.
        """
        img = Image.open(io.BytesIO(image))
        orig_width, orig_height = img.size
        orig_ratio = float(orig_height / orig_width)

        if orig_ratio >= ratio:
            width = orig_width
            height = int(width * ratio)
            x = 0
            y = (orig_height - height) / 2
        else:
            height = orig_height
            width = int(height / ratio)
            x = (orig_width - width) / 2
            y = 0

        regin = (x, y, width + x, height + y)

        new_img = img.crop(regin)

        # Make a PNG image from PIL without the need to create a temp
        # file.
        holder = io.BytesIO()
        new_img.save(holder, format='png')
        new_img_png = holder.getvalue()
        holder.close()

        return {'image': bytearray(new_img_png),
                'mimetype': 'image/png'}

    @classmethod
    def get_context(cls, records, header, data):
        context = super(ReportImageToolMixin, cls).get_context(
            records, header, data)

        context['image_crop'] = cls.image_crop
        context['image_resize'] = cls.image_resize
        context['image_keep_size'] = cls.image_keep_size

        return context



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
from trytond.model import ModelView, fields
from trytond.wizard import Wizard, StateView, StateAction, StateTransition, \
    Button
from trytond.transaction import Transaction

__all__ = ['OpenPediatricsGrowthChartsWHOReportStart',
           'OpenPediatricsGrowthChartsWHOReport']


class OpenPediatricsGrowthChartsWHOReportStart(ModelView):
    'Open Pediatrics Growth Charts WHO Report Start'
    __name__ = 'gnuhealth.pediatrics.growth.charts.who.report.open.start'

    indicator = fields.Selection([
        ('l/h-f-a', 'Length/height for age'),
        ('w-f-a', 'Weight for age'),
        ('bmi-f-a', 'Body mass index for age (BMI for age)'),
    ], 'Indicator', sort=False, required=True)
    measure = fields.Selection([
        ('p', 'percentiles'),
        ('z', 'z-scores'),
    ], 'Measure', required=True)


class OpenPediatricsGrowthChartsWHOReport(Wizard):
    'Open Pediatrics Growth Charts WHO Report'
    __name__ = 'gnuhealth.pediatrics.growth.charts.who.report.open'

    start = StateView(
        'gnuhealth.pediatrics.growth.charts.who.report.open.start',
        'health_pediatrics_growth_charts_who.'
        'growth_charts_who_open_start_view_form', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Open', 'choose', 'tryton-ok', default=True),
        ])
    choose = StateTransition()
    print_wfa = StateAction(
        'health_pediatrics_growth_charts_who.'
        'report_pediatrics_growth_charts_who_wfa')
    print_lhfa = StateAction(
        'health_pediatrics_growth_charts_who.'
        'report_pediatrics_growth_charts_who_lhfa')
    print_bmifa = StateAction(
        'health_pediatrics_growth_charts_who.'
        'report_pediatrics_growth_charts_who_bmifa')

    def transition_choose(self):
        if self.start.indicator == 'w-f-a':
            return 'print_wfa'
        elif self.start.indicator == 'l/h-f-a':
            return 'print_lhfa'
        else:
            return 'print_bmifa'

    def fill_data(self):
        return {
            'patient': Transaction().context.get('active_id'),
            'indicator': self.start.indicator,
            'measure': self.start.measure,
        }

    def do_print_wfa(self, action):
        return action, self.fill_data()

    def do_print_lhfa(self, action):
        return action, self.fill_data()

    def do_print_bmifa(self, action):
        return action, self.fill_data()



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH GYNECO package                         #
#                    health_gyneco.py: main module                      #
#########################################################################
import datetime
from dateutil.relativedelta import relativedelta
from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.pyson import Eval, Not, Bool, Equal
from trytond.pool import Pool, PoolMeta
from trytond.transaction import Transaction
from sql import Table
from sql.aggregate import Count
from trytond.modules.health.core import get_health_professional, \
    get_institution
from trytond.i18n import gettext

from .exceptions import PatientAlreadyPregnant


__all__ = [
    'PatientPregnancy', 'PregnancyResult', 'PrenatalEvaluation',
    'PuerperiumMonitor', 'Perinatal', 'PerinatalMonitor', 'GnuHealthPatient',
    'PatientMenstrualHistory', 'PatientMammographyHistory',
    'PatientPAPHistory', 'PatientColposcopyHistory']


class PatientPregnancy(ModelSQL, ModelView):
    'Patient Pregnancy'
    __name__ = 'gnuhealth.patient.pregnancy'

    # Show paient age at the moment of LMP
    def patient_age_at_pregnancy(self, name):
        if (self.patient.dob and self.lmp):
            rdelta = relativedelta(self.lmp, self.patient.dob)
            years = str(rdelta.years)
            return years
        else:
            return None

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient', required=True,
        domain=[('party.gender', '=', 'f')])

    gravida = fields.Integer(
        '#', required=True, help="Pregnancy number")

    computed_age = fields.Function(
        fields.Char(
            'Age',
            help="Computed patient age at the moment of LMP"),
        'patient_age_at_pregnancy')

    warning = fields.Boolean(
        'Warn', help='Check this box if this is pregancy'
        ' is or was NOT normal')
    warning_icon = fields.Function(fields.Char(
        'Pregnancy warning icon'), 'get_warn_icon')
    # reverse attribute deprecated in GH 5.0
    # Use the 'current_pregnancy' attribute for different states
    reverse = fields.Boolean(
        'Past', help="Past pregnancy.It will calculate the LMP "
        "from the delivery date and the gestational weeks. "
        "Set this field when the "
        "pregnancy information is referred by the patient, "
        "as a history taking procedure. Keep in mind "
        "that the reverse pregnancy data is subjective.",
    )
    reverse_weeks = fields.Integer(
        "Gest. Weeks", help="Number of weeks at "
        "the end of pregnancy.",
        states={
            'required': Not(Bool(Eval('current_pregnancy'))),
        }
    )

    lmp = fields.Date(
        'LMP', help="Last Menstrual Period",
        states={'readonly': Not(Bool(Eval('current_pregnancy'))),
                'required': Bool(Eval('current_pregnancy')),
                })

    pdd = fields.Function(
        fields.Date('Due Date', help='Pregnancy Due Date'),
        'get_pregnancy_data')

    prenatal_evaluations = fields.One2Many(
        'gnuhealth.patient.prenatal.evaluation', 'pregnancy',
        'Prenatal Evaluations')
    perinatal = fields.One2Many(
        'gnuhealth.perinatal', 'pregnancy', 'Perinatal Info')
    puerperium_monitor = fields.One2Many(
        'gnuhealth.puerperium.monitor',
        'pregnancy', 'Puerperium monitor')
    pregnancy_result = fields.One2Many(
        'gnuhealth.pregnancy.result', 'pregnancy', 'Result',
        states={
            'readonly': Bool(Eval('current_pregnancy')),
        })

    current_pregnancy = fields.Boolean(
        'Current',
        help='Set the field if the patient is currently pregnant')

    fetuses = fields.Integer('Fetuses', required=True)
    monozygotic = fields.Boolean('Monozygotic')
    pregnancy_end_result = fields.Selection([
        (None, ''),
        ('live_birth', 'Live birth'),
        ('abortion', 'Abortion'),
        ('stillbirth', 'Stillbirth'),
        ('status_unknown', 'Status unknown'),
    ], 'Result', sort=False,
        states={
            'required': Not(Bool(Eval('current_pregnancy'))),
    })
    pregnancy_end_date = fields.DateTime(
        'End',
        states={
            'readonly': Bool(Eval('current_pregnancy')),
            'required': Not(Bool(Eval('current_pregnancy'))),
        })
    bba = fields.Boolean(
        'BBA', help="Born Before Arrival",
        states={
            'readonly': Bool(Eval('current_pregnancy')),
        })
    home_birth = fields.Boolean(
        'Home Birth', help="Home Birth",
        states={
            'readonly': Bool(Eval('current_pregnancy')),
        })

    pregnancy_current_week = fields.Function(fields.Integer(
        'Week #', help='Current week'), 'get_pregnancy_data')
    iugr = fields.Selection([
        (None, ''),
        ('symmetric', 'Symmetric'),
        ('assymetric', 'Asymmetric'),
    ], 'IUGR', sort=False)

    institution = fields.Many2One(
        'gnuhealth.institution', 'Institution',
        help="Health center where this initial obstetric record was created")

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof', readonly=True,
        help="Health Professional who created this initial obstetric record")

    gravidae = fields.Function(
        fields.Integer(
            'Pregnancies',
            help="Number of pregnancies, computed from Obstetric history"),
        'patient_obstetric_info')
    premature = fields.Function(
        fields.Integer(
            'Premature',
            help="Preterm < 37 wks live births"), 'patient_obstetric_info')
    abortions = fields.Function(
        fields.Integer('Abortions'),
        'patient_obstetric_info')
    stillbirths = fields.Function(
        fields.Integer('Stillbirths'), 'patient_obstetric_info')

    blood_type = fields.Function(fields.Selection([
        (None, ''),
        ('A', 'A'),
        ('B', 'B'),
        ('AB', 'AB'),
        ('O', 'O'),
    ], 'Blood Type', sort=False),
        'patient_blood_info')

    rh = fields.Function(fields.Selection([
        (None, ''),
        ('+', '+'),
        ('-', '-'),
    ], 'Rh'),
        'patient_blood_info')

    hb = fields.Function(fields.Selection([
        (None, ''),
        ('aa', 'AA'),
        ('as', 'AS'),
        ('ss', 'SS'),
    ], 'Hb'),
        'patient_blood_info')

    notes = fields.Text("Notes")

    # Retrieve the info from the patient current GPA status
    def patient_obstetric_info(self, name):
        if (name == "gravidae"):
            return self.patient.gravida
        if (name == "premature"):
            return self.patient.premature
        if (name == "abortions"):
            return self.patient.abortions
        if (name == "stillbirths"):
            return self.patient.stillbirths

    # Retrieve Blood type and Rh and Hemoglobin
    def patient_blood_info(self, name):
        if (name == "blood_type"):
            return self.patient.blood_type
        if (name == "rh"):
            return self.patient.rh
        if (name == "hb"):
            return self.patient.hb

    # Show the values from patient upon entering the history
    @fields.depends(
        'patient',
        '_parent_patient.gravida',
        '_parent_patient.premature',
        '_parent_patient.abortions',
        '_parent_patient.stillbirths',
        '_parent_patient.blood_type',
        '_parent_patient.rh',
        '_parent_patient.hb')
    def on_change_patient(self):
        # Obsterics info
        self.gravidae = self.patient and self.patient.gravida
        self.premature = self.patient and self.patient.premature
        self.abortions = self.patient and self.patient.abortions
        self.stillbirths = self.patient and self.patient.stillbirths
        # Rh
        self.blood_type = self.patient and self.patient.blood_type
        self.rh = self.patient and self.patient.rh
        # Hb
        self.hb = self.patient and self.patient.hb

    @fields.depends('current_pregnancy', 'pregnancy_end_date', 'lmp')
    def on_change_pregnancy_end_date(self):
        '''Calculates the gestational weeks at the end of pregnancy
            based on the delivery date
            and the LMP when the setting the delivery date
            It works when LMP and pregnancy end date are set.
        '''
        if self.pregnancy_end_date and self.lmp:
            gestational_age = datetime.datetime.date(
                self.pregnancy_end_date) - self.lmp
            self.reverse_weeks = int((gestational_age.days) / 7)

    @classmethod
    def validate(cls, pregnancies):
        super(PatientPregnancy, cls).validate(pregnancies)
        for pregnancy in pregnancies:
            pregnancy.check_patient_current_pregnancy()

    def check_patient_current_pregnancy(self):
        ''' Check for only one current pregnancy in the patient '''
        pregnancy = Table('gnuhealth_patient_pregnancy')
        cursor = Transaction().connection.cursor()
        patient_id = int(self.patient.id)
        cursor.execute(*pregnancy.select(Count(pregnancy.patient),
                       where=(pregnancy.current_pregnancy == 'true') &
            (pregnancy.patient == patient_id)))

        records = cursor.fetchone()[0]
        if records > 1:
            raise PatientAlreadyPregnant(
                gettext('health_gyneco.msg_patient_already_pregnant'))

    @staticmethod
    def default_current_pregnancy():
        """ By default, GH will record the obstetric history of the
            patient. If we want to enter the information for the current
            pregnancy, we set the current_pregnancy field
        """
        return False

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @fields.depends('reverse_weeks', 'pregnancy_end_date')
    def on_change_with_lmp(self):
        '''
        Calculates the estimate on Last Menstrual Period
        using the reverse input method, taking the
        end of pregnancy date and number of weeks
        '''
        if (self.reverse_weeks and self.pregnancy_end_date):
            estimated_lmp = datetime.datetime.date(
                self.pregnancy_end_date -
                datetime.timedelta(self.reverse_weeks * 7))

            return estimated_lmp

    def get_pregnancy_data(self, name):
        """ Calculate the Pregnancy Due Date and the Number of
        weeks at the end of pregnancy when using the Last Menstrual
        Period parameter.
        It's not calculated when using the reverse input method
        """
        if (self.lmp):
            if name == 'pdd':
                return self.lmp + datetime.timedelta(days=280)
            if name == 'pregnancy_current_week':
                if self.current_pregnancy:
                    today = datetime.date.today()
                    weeks = int(((today - self.lmp).days) / 7)
                    return weeks
                if self.reverse_weeks:
                    return self.reverse_weeks

                    gestational_age = datetime.datetime.date(
                        self.pregnancy_end_date) - self.lmp
                    return int((gestational_age.days) / 7)
                else:
                    return 0

    def get_warn_icon(self, name):
        if self.warning:
            return 'gnuhealth-warning'

    @classmethod
    def search_rec_name(cls, name, clause):
        """ Include searching by the newborn
        """
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [bool_op,
                ('pregnancy_result.newborn',) + tuple(clause[1:]),
                ]

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to patient
        if (table_h.column_exist('name')
                and not table_h.column_exist('patient')):
            table_h.column_rename('name', 'patient')

        super().__register__(module)
        table_h = cls.__table_handler__(module)

    @classmethod
    def __setup__(cls):
        super(PatientPregnancy, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints += [
            ('gravida_uniq', Unique(t, t.patient, t.gravida),
             'This pregnancy code for this patient already exists'),
        ]
        cls._order.insert(0, ('lmp', 'DESC'))


class PrenatalEvaluation(ModelSQL, ModelView):
    'Prenatal and Antenatal Evaluations'
    __name__ = 'gnuhealth.patient.prenatal.evaluation'

    pregnancy = fields.Many2One(
        'gnuhealth.patient.pregnancy', 'Patient Pregnancy')
    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation',
        'Patient Evaluation', readonly=True)
    evaluation_date = fields.DateTime('Date', required=True)
    gestational_weeks = fields.Function(
        fields.Integer('Gestational Weeks'),
        'get_patient_evaluation_data')
    gestational_days = fields.Function(
        fields.Integer('Gestational days'),
        'get_patient_evaluation_data')
    hypertension = fields.Boolean(
        'Hypertension', help='Check this box if the'
        ' mother has hypertension')
    preeclampsia = fields.Boolean(
        'Preeclampsia', help='Check this box if the'
        ' mother has pre-eclampsia')
    overweight = fields.Boolean(
        'Overweight', help='Check this box if the'
        ' mother is overweight or obesity')
    diabetes = fields.Boolean(
        'Diabetes', help='Check this box if the mother'
        ' has glucose intolerance or diabetes')
    invasive_placentation = fields.Selection([
        (None, ''),
        ('normal', 'Normal decidua'),
        ('accreta', 'Accreta'),
        ('increta', 'Increta'),
        ('percreta', 'Percreta'),
    ], 'Placentation', sort=False)
    placenta_previa = fields.Boolean('Placenta Previa')
    vasa_previa = fields.Boolean('Vasa Previa')
    fundal_height = fields.Integer(
        'Fundal Height',
        help="Distance between the symphysis pubis and the uterine fundus "
        "(S-FD) in cm")
    fetus_heart_rate = fields.Integer(
        'Fetus heart rate', help='Fetus heart'
        ' rate')
    efw = fields.Integer('EFW', help="Estimated Fetal Weight")
    fetal_bpd = fields.Integer('BPD', help="Fetal Biparietal Diameter")
    fetal_ac = fields.Integer('AC', help="Fetal Abdominal Circumference")
    fetal_hc = fields.Integer('HC', help="Fetal Head Circumference")
    fetal_fl = fields.Integer('FL', help="Fetal Femur Length")
    oligohydramnios = fields.Boolean('Oligohydramnios')
    polihydramnios = fields.Boolean('Polihydramnios')
    iugr = fields.Boolean('IUGR', help="Intra Uterine Growth Restriction")

    urinary_activity_signs = fields.Boolean(
        "SUA",
        help="Signs of Urinary System Activity")

    digestive_activity_signs = fields.Boolean(
        "SDA",
        help="Signs of Digestive System Activity")

    notes = fields.Text("Notes")

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof', readonly=True,
        help="Health Professional in charge, or that who entered the "
        "information in the system.")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    def get_patient_evaluation_data(self, name):
        if name == 'gestational_weeks':
            gestational_age = datetime.datetime.date(self.evaluation_date) - \
                self.pregnancy.lmp
            return int((gestational_age.days) / 7)
        if name == 'gestational_days':
            gestational_age = datetime.datetime.date(self.evaluation_date) - \
                self.pregnancy.lmp
            return gestational_age.days

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to pregnancy
        if (table_h.column_exist('name')
                and not table_h.column_exist('pregnancy')):
            table_h.column_rename('name', 'pregnancy')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PuerperiumMonitor(ModelSQL, ModelView):
    'Puerperium Monitor'
    __name__ = 'gnuhealth.puerperium.monitor'

    pregnancy = fields.Many2One(
        'gnuhealth.patient.pregnancy', 'Patient Pregnancy')
    date = fields.DateTime('Date and Time', required=True)
    # Deprecated in 1.6.4 All the clinical information will be taken at the
    # main evaluation.
    # systolic / diastolic / frequency / temperature
    systolic = fields.Integer('Systolic Pressure')
    diastolic = fields.Integer('Diastolic Pressure')
    frequency = fields.Integer('Heart Frequency')
    temperature = fields.Float('Temperature')
    lochia_amount = fields.Selection([
        (None, ''),
        ('n', 'normal'),
        ('e', 'abundant'),
        ('h', 'hemorrhage'),
    ], 'Lochia amount', sort=False)
    lochia_color = fields.Selection([
        (None, ''),
        ('r', 'rubra'),
        ('s', 'serosa'),
        ('a', 'alba'),
    ], 'Lochia color', sort=False)
    lochia_odor = fields.Selection([
        (None, ''),
        ('n', 'normal'),
        ('o', 'offensive'),
    ], 'Lochia odor', sort=False)
    uterus_involution = fields.Integer(
        'Fundal Height',
        help="Distance between the symphysis pubis and the uterine fundus "
        "(S-FD) in cm")

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof', readonly=True,
        help="Health Professional in charge, or that who entered the "
        "information in the system.")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to pregnancy
        if (table_h.column_exist('name')
                and not table_h.column_exist('pregnancy')):
            table_h.column_rename('name', 'pregnancy')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class Perinatal(ModelSQL, ModelView):
    'Perinatal Information'
    __name__ = 'gnuhealth.perinatal'

    pregnancy = fields.Many2One(
        'gnuhealth.patient.pregnancy', 'Patient Pregnancy')
    admission_code = fields.Char('Code')
    # 1.6.4 Gravida number and abortion information go now in the pregnancy
    # header. It will be calculated as a function if needed
    gravida_number = fields.Integer('Gravida #')
    # Deprecated. Use End of Pregnancy as a general concept in the pregnancy
    # model
    # Abortion / Stillbirth / Live Birth
    abortion = fields.Boolean('Abortion')
    stillbirth = fields.Boolean('Stillbirth')
    admission_date = fields.DateTime(
        'Admission',
        help="Date when she was admitted to give birth", required=True)
    # Prenatal evaluations deprecated in 1.6.4. Will be computed automatically
    prenatal_evaluations = fields.Integer(
        'Prenatal evaluations',
        help="Number of visits to the doctor during pregnancy")
    start_labor_mode = fields.Selection([
        (None, ''),
        ('v', 'Vaginal - Spontaneous'),
        ('ve', 'Vaginal - Vacuum Extraction'),
        ('vf', 'Vaginal - Forceps Extraction'),
        ('c', 'C-section'),
    ], 'Delivery mode', sort=False)
    gestational_weeks = fields.Function(
        fields.Integer('Gestational wks'),
        'get_perinatal_information')
    gestational_days = fields.Integer('Days')
    fetus_presentation = fields.Selection([
        (None, ''),
        ('cephalic', 'Cephalic'),
        ('breech', 'Breech'),
        ('shoulder', 'Shoulder'),
    ], 'Fetus Presentation', sort=False)
    dystocia = fields.Boolean('Dystocia')
    placenta_incomplete = fields.Boolean(
        'Incomplete', help='Incomplete Placenta')
    placenta_retained = fields.Boolean('Retained', help='Retained Placenta')
    abruptio_placentae = fields.Boolean(
        'Abruptio Placentae',
        help='Abruptio Placentae')
    episiotomy = fields.Boolean('Episiotomy')
    # Vaginal tearing and forceps variables are deprecated in 1.6.4.
    # They are included in laceration and delivery mode respectively
    vaginal_tearing = fields.Boolean('Vaginal tearing')
    forceps = fields.Boolean('Forceps')
    monitoring = fields.One2Many(
        'gnuhealth.perinatal.monitor', 'perinatal',
        'Monitors')
    laceration = fields.Selection([
        (None, ''),
        ('perineal', 'Perineal'),
        ('vaginal', 'Vaginal'),
        ('cervical', 'Cervical'),
        ('broad_ligament', 'Broad Ligament'),
        ('vulvar', 'Vulvar'),
        ('rectal', 'Rectal'),
        ('bladder', 'Bladder'),
        ('urethral', 'Urethral'),
    ], 'Lacerations', sort=False)
    hematoma = fields.Selection([
        (None, ''),
        ('vaginal', 'Vaginal'),
        ('vulvar', 'Vulvar'),
        ('retroperitoneal', 'Retroperitoneal'),
    ], 'Hematoma', sort=False)
    notes = fields.Text('Notes')

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof', readonly=True,
        help="Health Professional in charge, or that who entered the "
        "information in the system.")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    def get_perinatal_information(self, name):
        if name == 'gestational_weeks':
            gestational_age = datetime.datetime.date(self.admission_date) - \
                self.pregnancy.lmp
            return int((gestational_age.days) / 7)

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to pregnancy
        if (table_h.column_exist('name')
                and not table_h.column_exist('pregnancy')):
            table_h.column_rename('name', 'pregnancy')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PerinatalMonitor(ModelSQL, ModelView):
    'Perinatal Monitor'
    __name__ = 'gnuhealth.perinatal.monitor'

    perinatal = fields.Many2One(
        'gnuhealth.perinatal',
        'Patient Perinatal Evaluation')
    date = fields.DateTime('Date and Time')
    systolic = fields.Integer('Systolic Pressure')
    diastolic = fields.Integer('Diastolic Pressure')
    contractions = fields.Integer('Contractions')
    frequency = fields.Integer('Mother\'s Heart Frequency')
    dilation = fields.Integer('Cervix dilation')
    f_frequency = fields.Integer('Fetus Heart Frequency')
    meconium = fields.Boolean('Meconium')
    bleeding = fields.Boolean('Bleeding')
    fundal_height = fields.Integer('Fundal Height')
    fetus_position = fields.Selection([
        (None, ''),
        ('o', 'Occiput / Cephalic Posterior'),
        ('fb', 'Frank Breech'),
        ('cb', 'Complete Breech'),
        ('t', 'Transverse Lie'),
        ('flb', 'Footling Breech'),
    ], 'Fetus Position', sort=False)

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to perinatal
        if (table_h.column_exist('name')
                and not table_h.column_exist('perinatal')):
            table_h.column_rename('name', 'perinatal')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PregnancyResult(ModelSQL, ModelView):
    'Pregnancy Result'
    __name__ = 'gnuhealth.pregnancy.result'

    pregnancy = fields.Many2One(
        'gnuhealth.patient.pregnancy', 'Patient Pregnancy')

    result = fields.Selection([
        (None, ''),
        ('live_birth', 'Live birth'),
        ('abortion', 'Abortion'),
        ('stillbirth', 'Stillbirth'),
        ('status_unknown', 'Status unknown'),
    ], 'Result', sort=False)

    newborn = fields.Many2One(
        'party.party', 'Newborn',
        domain=[
            ('is_person', '=', True),
            ('dob', '=', Eval('dob')),
            ],
        depends=['pregnancy', 'dob'])

    dob = fields.Function(fields.Date('Date of birth'), 'get_dob')
    delivery_mode = fields.Selection([
        (None, ''),
        ('v', 'Vaginal - Spontaneous'),
        ('ve', 'Vaginal - Vacuum Extraction'),
        ('vf', 'Vaginal - Forceps Extraction'),
        ('c', 'C-section'),
    ], 'Delivery mode', sort=False)

    labor_time = fields.Integer(
        'Labor time', help="Total labor time in hours passive + active")

    short_comment = fields.Char(
        'Comments', help="Short extra information")

    def get_dob(self, name):
        if self.pregnancy and self.pregnancy.pregnancy_end_date:
            return self.pregnancy.pregnancy_end_date.date()

    # Retrieve date of birth upon entering the pregnancy
    @fields.depends(
        'pregnancy', '_parent_pregnancy.pregnancy_end_date')
    def on_change_pregnancy(self):
        if self.pregnancy and self.pregnancy.pregnancy_end_date:
            return self.pregnancy.pregnancy_end_date.date()

    # Get the baby date of birth from pregnancy end date
    @fields.depends(
        'result', 'pregnancy', '_parent_pregnancy.pregnancy_end_date')
    def on_change_result(self):
        if (self.pregnancy):
            self.dob = self.pregnancy.pregnancy_end_date.date()

    @classmethod
    def __setup__(cls):
        super(PregnancyResult, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints += [
            ('newborn_uniq', Unique(t, t.newborn),
             'Our records show that the newborn is from another pregnancy'),
        ]


class GnuHealthPatient(metaclass=PoolMeta):

    """Add to the Medical patient_data class (gnuhealth.patient) the
      gynecological and obstetric fields.
    """
    __name__ = 'gnuhealth.patient'

    currently_pregnant = fields.Function(
        fields.Boolean('Pregnant'),
        'get_pregnancy_info')
    fertile = fields.Boolean(
        'Fertile',
        help="Check if patient is in fertile age")
    menarche = fields.Integer('Menarche age')
    menopausal = fields.Boolean('Menopausal')
    menopause = fields.Integer('Menopause age')
    mammography = fields.Boolean(
        'Mammography',
        help="Check if the patient does periodic mammographys")
    mammography_last = fields.Date(
        'Last mammography',
        help="Enter the date of the last mammography")
    breast_self_examination = fields.Boolean(
        'Breast self-examination',
        help="Check if patient does and knows how to self examine her breasts")
    pap_test = fields.Boolean(
        'PAP test',
        help="Check if patient does periodic cytologic pelvic smear screening")
    pap_test_last = fields.Date(
        'Last PAP test',
        help="Enter the date of the last Papanicolau test")
    colposcopy = fields.Boolean(
        'Colposcopy',
        help="Check if the patient has done a colposcopy exam")
    colposcopy_last = fields.Date(
        'Last colposcopy',
        help="Enter the date of the last colposcopy")
    # From version 2.6 Gravida, premature, abortions and stillbirths are now
    # functional fields, computed from the obstetric information
    gravida = fields.Function(fields.Integer(
        'Pregnancies',
        help="Number of pregnancies, computed from Obstetric history"),
        'patient_obstetric_info')
    premature = fields.Function(fields.Integer(
        'Premature',
        help="Preterm < 37 wks live births"), 'patient_obstetric_info')

    abortions = fields.Function(
        fields.Integer('Abortions'),
        'patient_obstetric_info')
    stillbirths = fields.Function(
        fields.Integer('Stillbirths'),
        'patient_obstetric_info')

    full_term = fields.Integer('Full Term', help="Full term pregnancies")
    # GPA Deprecated in 1.6.4. It will be used as a function or report from the
    # other fields
    #    gpa = fields.Char('GPA',
    #        help="Gravida, Para, Abortus Notation. For example G4P3A1 : 4 "
    #        "Pregnancies, 3 viable and 1 abortion")
    # Deprecated. The born alive number will be calculated from pregnancies -
    # abortions - stillbirths
    #    born_alive = fields.Integer('Born Alive')
    # Deceased in 1st week or after 2nd weeks are deprecated since 1.6.4 .
    # The information will be retrieved from the neonatal or infant record
    #    deaths_1st_week = fields.Integer('Deceased during 1st week',
    #        help="Number of babies that die in the first week")
    #    deaths_2nd_week = fields.Integer('Deceased after 2nd week',
    #        help="Number of babies that die after the second week")
    # Perinatal Deprecated since 1.6.4 - Included in the obstetric history
    #    perinatal = fields.One2Many('gnuhealth.perinatal', 'name',
    #        'Perinatal Info')
    menstrual_history = fields.One2Many(
        'gnuhealth.patient.menstrual_history',
        'patient', 'Menstrual History')
    mammography_history = fields.One2Many(
        'gnuhealth.patient.mammography_history',
        'patient', 'Mammography History',
        states={'invisible': Not(Bool(Eval('mammography')))},
    )
    pap_history = fields.One2Many(
        'gnuhealth.patient.pap_history', 'patient',
        'PAP smear History',
        states={'invisible': Not(Bool(Eval('pap_test')))},
    )
    colposcopy_history = fields.One2Many(
        'gnuhealth.patient.colposcopy_history', 'patient',
        'Colposcopy History',
        states={'invisible': Not(Bool(Eval('colposcopy')))},
    )
    pregnancy_history = fields.One2Many(
        'gnuhealth.patient.pregnancy', 'patient',
        'Pregnancies')

    def get_pregnancy_info(self, name):
        if name == 'currently_pregnant':
            for pregnancy_history in self.pregnancy_history:
                if pregnancy_history.current_pregnancy:
                    return True
        return False

    def patient_obstetric_info(self, name):
        ''' Return the number of pregnancies, perterm,
        abortion and stillbirths '''

        counter = 0
        pregnancies = len(self.pregnancy_history)
        if (name == "gravida"):
            return pregnancies

        if (name == "premature"):
            prematures = 0
            while counter < pregnancies:
                result = self.pregnancy_history[counter].pregnancy_end_result
                preg_weeks = self.pregnancy_history[counter].reverse_weeks
                if (result == "live_birth" and preg_weeks):
                    if preg_weeks < 37:
                        prematures = prematures + 1
                counter = counter + 1
            return prematures

        if (name == "abortions"):
            abortions = 0
            while counter < pregnancies:
                result = self.pregnancy_history[counter].pregnancy_end_result
                if (result == "abortion"):
                    abortions = abortions + 1
                counter = counter + 1

            return abortions

        if (name == "stillbirths"):
            stillbirths = 0
            while counter < pregnancies:
                result = self.pregnancy_history[counter].pregnancy_end_result
                if (result == "stillbirth"):
                    stillbirths = stillbirths + 1
                counter = counter + 1
            return stillbirths

    @classmethod
    def view_attributes(cls):
        return super(GnuHealthPatient, cls).view_attributes() + [
            ('//page[@id="page_gyneco_obs"]', 'states', {
                'invisible': Equal(Eval('biological_sex'), 'm'),
            })]


class PatientMenstrualHistory(ModelSQL, ModelView):
    'Menstrual History'
    __name__ = 'gnuhealth.patient.menstrual_history'

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient', readonly=True, required=True)
    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation', 'Evaluation',
        domain=[('patient', '=', Eval('patient'))],
        depends=['patient'])
    evaluation_date = fields.Date(
        'Date', help="Evaluation Date",
        required=True)
    lmp = fields.Date('LMP', help="Last Menstrual Period", required=True)
    lmp_length = fields.Integer('Length', required=True)
    is_regular = fields.Boolean('Regular')
    dysmenorrhea = fields.Boolean('Dysmenorrhea')
    frequency = fields.Selection([
        ('amenorrhea', 'amenorrhea'),
        ('oligomenorrhea', 'oligomenorrhea'),
        ('eumenorrhea', 'eumenorrhea'),
        ('polymenorrhea', 'polymenorrhea'),
    ], 'frequency', sort=False)
    volume = fields.Selection([
        ('hypomenorrhea', 'hypomenorrhea'),
        ('normal', 'normal'),
        ('menorrhagia', 'menorrhagia'),
    ], 'volume', sort=False)

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Reviewed', readonly=True,
        help="Health Professional who reviewed the information")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @staticmethod
    def default_evaluation_date():
        return Pool().get('ir.date').today()

    @staticmethod
    def default_frequency():
        return 'eumenorrhea'

    @staticmethod
    def default_volume():
        return 'normal'

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to patient
        if (table_h.column_exist('name')
                and not table_h.column_exist('patient')):
            table_h.column_rename('name', 'patient')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PatientMammographyHistory(ModelSQL, ModelView):
    'Mammography History'
    __name__ = 'gnuhealth.patient.mammography_history'

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient', readonly=True, required=True)
    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation', 'Evaluation',
        domain=[('patient', '=', Eval('patient'))],
        depends=['patient'])
    evaluation_date = fields.Date('Date', help="Date", required=True)
    last_mammography = fields.Date('Previous', help="Last Mammography")
    result = fields.Selection([
        (None, ''),
        ('normal', 'normal'),
        ('abnormal', 'abnormal'),
    ], 'result',
        help="Please check the lab test results if the module is "
        "installed", sort=False)
    comments = fields.Char('Remarks')

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Reviewed', readonly=True,
        help="Health Professional who last reviewed the test")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @staticmethod
    def default_evaluation_date():
        return Pool().get('ir.date').today()

    @staticmethod
    def default_last_mammography():
        return Pool().get('ir.date').today()

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to patient
        if (table_h.column_exist('name')
                and not table_h.column_exist('patient')):
            table_h.column_rename('name', 'patient')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PatientPAPHistory(ModelSQL, ModelView):
    'PAP Test History'
    __name__ = 'gnuhealth.patient.pap_history'

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient', readonly=True, required=True)
    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation', 'Evaluation',
        domain=[('patient', '=', Eval('patient'))],
        depends=['patient'])
    evaluation_date = fields.Date('Date', help="Date", required=True)
    last_pap = fields.Date('Previous', help="Last Papanicolau")
    result = fields.Selection([
        (None, ''),
        ('negative', 'Negative'),
        ('c1', 'ASC-US'),
        ('c2', 'ASC-H'),
        ('g1', 'ASG'),
        ('c3', 'LSIL'),
        ('c4', 'HSIL'),
        ('g4', 'AIS'),
    ], 'result', help="Please check the lab results if the module is "
        "installed", sort=False)
    comments = fields.Char('Remarks')

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Reviewed', readonly=True,
        help="Health Professional who last reviewed the test")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @staticmethod
    def default_evaluation_date():
        return Pool().get('ir.date').today()

    @staticmethod
    def default_last_pap():
        return Pool().get('ir.date').today()

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to patient
        if (table_h.column_exist('name')
                and not table_h.column_exist('patient')):
            table_h.column_rename('name', 'patient')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PatientColposcopyHistory(ModelSQL, ModelView):
    'Colposcopy History'
    __name__ = 'gnuhealth.patient.colposcopy_history'

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient', readonly=True, required=True)
    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation', 'Evaluation',
        domain=[('patient', '=', Eval('patient'))],
        depends=['patient'])
    evaluation_date = fields.Date('Date', help="Date", required=True)
    last_colposcopy = fields.Date('Previous', help="Last colposcopy")
    result = fields.Selection([
        (None, ''),
        ('normal', 'normal'),
        ('abnormal', 'abnormal'),
    ], 'result',
        help="Please check the lab test results if the module is "
        "installed", sort=False)
    comments = fields.Char('Remarks')

    institution = fields.Many2One('gnuhealth.institution', 'Institution')

    healthprof = fields.Many2One(
        'gnuhealth.healthprofessional', 'Reviewed', readonly=True,
        help="Health Professional who last reviewed the test")

    @staticmethod
    def default_institution():
        return get_institution()

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @staticmethod
    def default_evaluation_date():
        return Pool().get('ir.date').today()

    @staticmethod
    def default_last_colposcopy():
        return Pool().get('ir.date').today()

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to patient
        if (table_h.column_exist('name')
                and not table_h.column_exist('patient')):
            table_h.column_rename('name', 'patient')

        super().__register__(module)
        table_h = cls.__table_handler__(module)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH QR_CODES PACKAGE                       #
#                      health_qrcodes.py: Main module                   #
#########################################################################
import qrcode
import barcode
import io
from trytond.model import fields
from trytond.pool import PoolMeta


__all__ = ['Party', 'Patient', 'Appointment', 'Newborn', 'LabTest']


# Remove text of barcode.
from barcode.base import Barcode
Barcode.default_writer_options['write_text'] = False


class Party(metaclass=PoolMeta):
    __name__ = 'party.party'

    # Add the CODE39 Code to the Person for ID purposes
    barcode = fields.Function(fields.Binary('Code39'), 'make_code39')

    def make_code39(self, name):
        # Create the Code39 bar code to encode the Person ID
        party_puid = self.ref or ''
        puid = f'{party_puid}'

        CODE39 = barcode.get_barcode_class('code39')

        code39 = CODE39(puid, add_checksum=False)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        code39.write(holder)
        code39_png = holder.getvalue()
        holder.close()

        return bytearray(code39_png)


# Add the QR field and QR image in the patient model
class Patient(metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient'

    # Add the QR Code to the Patient
    qr = fields.Function(fields.Binary('QR Code'), 'make_qrcode')
    barcode = fields.Function(fields.Binary('Code39'), 'make_code39')

    def make_qrcode(self, name):
        # Create the QR code

        patient_puid = self.puid or ''
        patient_blood_type = self.blood_type or ''
        patient_rh = self.rh or ''
        patient_gender = self.party.gender_str or ''
        patient_dob = ''

        if (self.dob):
            patient_dob = str(self.dob)

        qr_string = f'{patient_puid}\n' \
            f'Name: {self.party.rec_name}\n' \
            f'Gender: {patient_gender}\n' \
            f'DoB: {patient_dob}\n' \
            f'Blood Type: {patient_blood_type} {patient_rh}'

        qr_image = qrcode.make(qr_string)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        qr_image.save(holder)
        qr_png = holder.getvalue()
        holder.close()

        return bytearray(qr_png)

    def make_code39(self, name):
        # Create the Code39 bar code to encode the patient's PUID

        patient_puid = self.puid or ''
        puid = f'{patient_puid}'

        CODE39 = barcode.get_barcode_class('code39')

        code39 = CODE39(puid, add_checksum=False)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        code39.write(holder)
        code39_png = holder.getvalue()
        holder.close()

        return bytearray(code39_png)


class Appointment(metaclass=PoolMeta):
    __name__ = 'gnuhealth.appointment'

    # Add the QR Code to the Appointment
    qr = fields.Function(fields.Binary('QR Code'), 'make_qrcode')
    barcode = fields.Function(fields.Binary('Code39'), 'make_code39')

    def make_qrcode(self, name):
        # Create the QR code

        appointment_healthprof = ''
        appointment_patient = ''
        patient_puid = ''
        appointment_specialty = ''
        appointment_date = ''
        appointment = ''

        if (self.name):
            appointment = self.name

        if (self.healthprof):
            appointment_healthprof = str(self.healthprof.rec_name) or ''

        if (self.patient):
            appointment_patient = self.patient.rec_name or ''
            patient_puid = self.patient and self.patient.puid

        if (self.appointment_date):
            appointment_date = str(self.appointment_date)

        if (self.speciality):
            appointment_specialty = str(self.speciality.rec_name) or ''

        qr_string = f'{appointment}\n' \
            f'Name: {appointment_patient}\n' \
            f'PUID: {patient_puid}\n' \
            f'Specialty: {appointment_specialty}\n' \
            f'Health Prof: {appointment_healthprof}\n' \
            f'Date: {appointment_date}'

        qr_image = qrcode.make(qr_string)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        qr_image.save(holder)
        qr_png = holder.getvalue()
        holder.close()

        return bytearray(qr_png)

    def make_code39(self, name):
        if (self.name):
            appointment = f'{self.name}'
        else:
            # Use ' ' instead '', for the barcode of '' can not be
            # recognized by the barcode scanner.
            appointment = ' '

        CODE39 = barcode.get_barcode_class('code39')

        code39 = CODE39(appointment, add_checksum=False)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        code39.write(holder)
        code39_png = holder.getvalue()
        holder.close()

        return bytearray(code39_png)


class Newborn(metaclass=PoolMeta):
    'NewBorn'
    __name__ = 'gnuhealth.newborn'

    # Add the QR Code to the Newborn
    qr = fields.Function(fields.Binary('QR Code'), 'make_qrcode')

    def make_qrcode(self, name):
        # Create the QR code

        newborn_mother_name = self.mother and self.mother.rec_name or ''
        newborn_mother_id = self.mother and self.mother.puid or ''

        newborn_name = self.newborn_name or self.patient.rec_name or ''
        newborn_id = self.patient.puid or ''
        newborn_sex = self.sex_str or ''
        newborn_birth_date = self.birth_date or ''

        qr_string = f'{newborn_id}\n' \
            f'Name: {newborn_name}\n' \
            f'Mother: {newborn_mother_name}\n' \
            f'Mother\'s PUID: {newborn_mother_id}\n' \
            f'Sex: {newborn_sex}\n' \
            f'DoB: {str(newborn_birth_date)}'

        qr_image = qrcode.make(qr_string)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        qr_image.save(holder)
        qr_png = holder.getvalue()
        holder.close()

        return bytearray(qr_png)


class LabTest(metaclass=PoolMeta):
    __name__ = 'gnuhealth.lab'

    # Add the QR Code to the Lab Test
    qr = fields.Function(fields.Binary('QR Code'), 'make_qrcode')
    bar = fields.Function(fields.Binary('Bar Code39'), 'make_barcode')

    def make_qrcode(self, name):
        # Create the QR code

        labtest_id = self.name or ''
        labtest_type = self.test or ''

        patient_puid = self.patient and self.patient.puid or ''
        patient_name = self.patient and self.patient.rec_name or ''
        source = self.other_source

        requestor_name = self.requestor and self.requestor.rec_name or ''

        if self.is_patient():
            qr_string = f'{labtest_id}\n' \
                f'Test: {labtest_type.rec_name}\n' \
                f'PUID: {patient_puid}\n' \
                f'Patient: {patient_name}\n' \
                f'Requestor: {requestor_name}'
        else:
            qr_string = f'{labtest_id}\n' \
                f'Test: {labtest_type.rec_name}\n' \
                f'Source: {source}\n' \
                f'Requestor: {requestor_name}'

        qr_image = qrcode.make(qr_string)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        qr_image.save(holder)
        qr_png = holder.getvalue()
        holder.close()

        return bytearray(qr_png)

    def make_barcode(self, name):
        # Create the Code39 bar code to encode the TEST ID

        labtest_id = self.name or ''

        CODE39 = barcode.get_barcode_class('code39')

        code39 = CODE39(labtest_id, add_checksum=False)

        # Make a PNG image from PIL without the need to create a temp file

        holder = io.BytesIO()
        code39.write(holder)
        code39_png = holder.getvalue()
        holder.close()

        return bytearray(code39_png)



# SPDX-FileCopyrightText: 2020-2021 National University of Entre Rios (UNER)
#                         School of Engineering
#                         <saludpublica@ingenieria.uner.edu.ar>
# SPDX-FileCopyrightText: 2020 Mario Puntin <mario@silix.com.ar>
# SPDX-FileCopyrightText: 2020-2024 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2020-2024 GNU Solidario <health@gnusolidario.org>

# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH DENTISTRY package                          #
#                 health_dentisty.py: main module                       #
#########################################################################
import json
from datetime import date
from collections import defaultdict

from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.pyson import Eval, Equal
from trytond.pool import PoolMeta

from trytond.modules.health.core import get_health_professional

__all__ = ['PatientData', 'DentistryTreatment', 'DentistryProcedure',
           'TreatmentProcedure']


TOOTH_STATE = [
    ('', ''),
    ('D', 'Decayed'),
    ('M', 'Missing'),
    ('F', 'Filled'),
    ('E', 'For Extraction'),
]

STATE_LEGENDS = {
    '': '      ',
    'D': '  *  ',
    'M': '  X  ',
    'F': '  #  ',
    'E': '  =  '
}

TEETH = [
    ('11', '11'), ('12', '12'), ('13', '13'), ('14', '14'),
    ('15', '15'), ('16', '16'), ('17', '17'), ('18', '18'),
    ('21', '21'), ('22', '22'), ('23', '23'), ('24', '24'),
    ('25', '25'), ('26', '26'), ('27', '27'), ('28', '28'),
    ('31', '31'), ('32', '32'), ('33', '33'), ('34', '34'),
    ('35', '35'), ('36', '36'), ('37', '37'), ('38', '38'),
    ('41', '41'), ('42', '42'), ('43', '43'), ('44', '44'),
    ('45', '45'), ('46', '46'), ('47', '47'), ('48', '48'),
    ('51', '51'), ('52', '52'), ('53', '53'), ('54', '54'), ('55', '55'),
    ('61', '61'), ('62', '62'), ('63', '63'), ('64', '64'), ('65', '65'),
    ('71', '71'), ('72', '72'), ('73', '73'), ('74', '74'), ('75', '75'),
    ('81', '81'), ('82', '82'), ('83', '83'), ('84', '84'), ('85', '85'),
]

TREATMENT_TEETH = TEETH + [(None, '')]


class PatientData (metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient'

    dentistry_treatments = fields.One2Many(
        'gnuhealth.dentistry.treatment',
        'patient', 'Treatment', readonly=True)
    dental_schema = fields.Text('Dental Schema')
    dental_schema_primary = fields.Text('Primary Schema')
    teeth1 = fields.Function(fields.Char('Quadrant 1'), 'get_status')
    teeth2 = fields.Function(fields.Char('Quadrant 2'), 'get_status')
    teeth3 = fields.Function(fields.Char('Quadrant 3'), 'get_status')
    teeth4 = fields.Function(fields.Char('Quadrant 4'), 'get_status')
    use_primary_schema = fields.Boolean('Primary Schema',
                                        help='Use Primary Schema')
    teeth5 = fields.Function(fields.Char('Quadrant 5'), 'get_status_primary')
    teeth6 = fields.Function(fields.Char('Quadrant 6'), 'get_status_primary')
    teeth7 = fields.Function(fields.Char('Quadrant 7'), 'get_status_primary')
    teeth8 = fields.Function(fields.Char('Quadrant 8'), 'get_status_primary')
    dmft_index = fields.Function(fields.Integer('DMFT Index'),
                                 'get_dmft_index')
    dmft_index_primary = fields.Function(
        fields.Integer('dmft index', help='dmft index for primary teeth',
                       states={'invisible': ~Eval('use_primary_schema')}),
        'get_dmft_index_primary')

    @classmethod
    def __setup__(cls):
        super(PatientData, cls).__setup__()
        cls._buttons.update({
            'set_odontogram_wizard': {
                'readonly': Eval('deceased'),
            },
        })

    @classmethod
    def view_attributes(cls):
        attributes = super(PatientData, cls).view_attributes()
        attributes.extend([
            ('///group[@id="dental_main_info"]/'
                'group[@id="dental_schema_primary"]',
             'states', {'invisible': ~Eval('use_primary_schema')})
        ])
        return attributes

    @staticmethod
    def default_use_primary_schema():
        return False

    @fields.depends('dental_schema')
    def get_status(self, name=None):
        if not self.dental_schema:
            return ''
        quad = name[5:]
        teeth = json.loads(self.dental_schema)
        res = ''
        loop = list(range(1, 9)) if quad in ['2', '3'] \
            else list(range(8, 0, -1))
        for i in loop:
            res += ' ' * 6 + quad + str(i) + ':' + \
                STATE_LEGENDS[teeth[quad + str(i)]['ts']]
        return res

    @fields.depends('dental_schema_primary')
    def get_status_primary(self, name=None):
        if not self.dental_schema_primary:
            return ''
        quad = name[5:]
        teeth = json.loads(self.dental_schema_primary)
        res = ' ' * 24
        loop = list(range(1, 6)) if quad in ['6', '7'] \
            else list(range(5, 0, -1))
        for i in loop:
            res += ' ' * 6 + quad + str(i) + ':' + \
                STATE_LEGENDS[teeth[quad + str(i)]['ts']]
        return res

    @fields.depends('dental_schema')
    def get_dmft_index(self, name=None):
        if not self.dental_schema:
            return None
        teeth = json.loads(self.dental_schema)
        res = 0
        quadrant = ['1', '2', '3', '4']
        tooth = ['1', '2', '3', '4', '5', '6', '7', '8']
        for q in quadrant:
            for t in tooth:
                if teeth[q + t]['ts'] in ['D', 'M', 'F']:
                    res += 1
        return res

    @fields.depends('dental_schema_primary')
    def get_dmft_index_primary(self, name=None):
        if not self.dental_schema_primary:
            return None
        teeth = json.loads(self.dental_schema_primary)
        res = 0
        quadrant = ['5', '6', '7', '8']
        tooth = ['1', '2', '3', '4', '5']
        for q in quadrant:
            for t in tooth:
                if teeth[q + t]['ts'] in ['D', 'M', 'F']:
                    res += 1
        return res

    @classmethod
    @ModelView.button_action('health_dentistry.wizard_set_odontogram')
    def set_odontogram_wizard(cls, patients):
        pass


class DentistryTreatment(ModelSQL, ModelView):
    'Dentistry Treatment'
    __name__ = 'gnuhealth.dentistry.treatment'

    STATES = {'readonly': Eval('state') == 'done'}

    patient = fields.Many2One('gnuhealth.patient', 'Patient', required=True)
    treatment_date = fields.Date('Date', states=STATES)
    healthprof = fields.Many2One('gnuhealth.healthprofessional',
                                 'Health Prof', help="Health professional",
                                 states=STATES)
    procedures = fields.One2Many('gnuhealth.dentistry.treatment.procedure',
                                 'treatment', 'Procedures', states=STATES)
    notes = fields.Text('Notes', help="Extra Information", states=STATES)
    procedures_info = fields.Function(fields.Char("Procedures info"),
                                      'get_procedures_info')
    signed_by = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Signed by', readonly=True,
        states={'invisible': Equal(Eval('state'),
                                   'pending')},
        help="Health Professional that finished the treatment")
    state = fields.Selection([
        ('pending', 'Pending'),
        ('done', 'Done'),
    ], 'State', readonly=True, sort=False)
    state_string = state.translated('state')

    @classmethod
    def __setup__(cls):
        super(DentistryTreatment, cls).__setup__()
        cls._order.insert(0, ('id', 'DESC'))
        cls._buttons.update({
            'load_procedure': {
                'readonly': ~Eval('state').in_(['pending']),
            },
            'set_odontogram': {
                'readonly': ~Eval('state').in_(['pending']),
            },
            'end_treatment': {
                'invisible': Equal(Eval('state'), 'done')
            },
        })

    @classmethod
    @ModelView.button_action('health_dentistry.load_procedure')
    def load_procedure(cls, treatments):
        pass

    @classmethod
    @ModelView.button_action(
        'health_dentistry.wizard_set_odontogram_from_treatment')
    def set_odontogram(cls, treatments):
        pass

    @staticmethod
    def default_treatment_date():
        return date.today()

    @staticmethod
    def default_state():
        return 'pending'

    @staticmethod
    def default_healthprof():
        return get_health_professional()

    @classmethod
    def get_procedures_info(cls, treatments, names):
        result = {
            'procedures_info': {}
        }
        for t in treatments:
            data = defaultdict(list)
            info = []
            for p in t.procedures:
                data[p.procedure.code].append(p.tooth if p.tooth else '')
            for k, v in data.items():
                info.append(k + ' (' + ', '.join(v) + ')')
            result['procedures_info'][t.id] = ', '.join(info)
        return result

    @classmethod
    @ModelView.button
    def end_treatment(cls, treatments):
        signing_hp = get_health_professional()
        cls.write(treatments, {
            'state': 'done',
            'signed_by': signing_hp,
        })


class DentistryProcedure(ModelSQL, ModelView):
    'Dentistry Procedure'
    __name__ = 'gnuhealth.dentistry.procedure'

    name = fields.Char('Procedure', required=True, translate=True)
    code = fields.Char('Code', required=True,
                       help='Please use CAPITAL LETTERS and no spaces')

    @classmethod
    def __setup__(cls):
        t = cls.__table__()
        cls._sql_constraints = [
            ('name_uniq', Unique(t, t.name),
                'The Procedure must be unique !'),
            ('code_uniq', Unique(t, t.code),
                'The CODE must be unique !'),
        ]
        super(DentistryProcedure, cls).__setup__()


class TreatmentProcedure(ModelSQL, ModelView):
    'Treatment Procedure'
    __name__ = 'gnuhealth.dentistry.treatment.procedure'

    treatment = fields.Many2One('gnuhealth.dentistry.treatment', 'Treatment',
                                required=True)
    tooth = fields.Selection(TREATMENT_TEETH, 'Tooth')
    procedure = fields.Many2One('gnuhealth.dentistry.procedure', 'Procedure',
                                required=True)
    root = fields.Boolean('Root')
    occlusal = fields.Boolean('Occlusal')
    vestibular = fields.Boolean('Vestibular')
    lingual = fields.Boolean('Lingual')
    mesial = fields.Boolean('Mesial')
    distal = fields.Boolean('Distal')

    @classmethod
    def __setup__(cls):
        super(TreatmentProcedure, cls).__setup__()
        cls._order.insert(0, ('procedure', 'ASC'))
        cls._order.insert(1, ('tooth', 'ASC'))



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2011-2014 Sebastian Marro <smarro@thymbra.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH CALENDAR PACKAGE                         #
#                health_calendar.py: Main calendar module               #
#########################################################################
from trytond.model import fields
from trytond.pyson import Eval, Not, Bool
from trytond.pool import Pool, PoolMeta
from trytond.i18n import gettext

from .exceptions import (AppointmentEndDateBeforeStart)

__all__ = ['User', 'Appointment']


class User(metaclass=PoolMeta):
    __name__ = "res.user"

    use_caldav = fields.Boolean('CalDAV', help="The user has a CalDav account")
    calendar = fields.Many2One(
        'calendar.calendar', 'Calendar',
        states={
            'invisible': Not(Bool(Eval('use_caldav'))),
            'required': Bool(Eval('use_caldav')),
        })


class Appointment(metaclass=PoolMeta):
    __name__ = 'gnuhealth.appointment'

    event = fields.Many2One(
        'calendar.event', 'CalDAV Event', readonly=True,
        help="Calendar Event",
        states={'invisible': Not(Bool(Eval('event')))})
    appointment_date_end = fields.DateTime(
        'End', help="Appointment end date and time.")

    @classmethod
    def validate(cls, appointments):
        super(Appointment, cls).validate(appointments)
        for appointment in appointments:
            appointment.validate_appointment_dates()

    def validate_appointment_dates(self):
        if self.appointment_date_end:
            if (self.appointment_date_end < self.appointment_date):
                raise AppointmentEndDateBeforeStart(gettext(
                    'health_calendar.msg_appointment_end_date_before_start',
                    appointment_date=self.appointment_date,
                    appointment_date_end=self.appointment_date_end
                )
                )

    @classmethod
    def create(cls, vlist):
        pool = Pool()
        Event = pool.get('calendar.event')
        Patient = pool.get('gnuhealth.patient')
        Healthprof = pool.get('gnuhealth.healthprofessional')

        vlist = [x.copy() for x in vlist]

        for values in vlist:
            # Check if the event exists (eg, created from a caldav client)
            # If so, then do not create it again.
            if 'event' in values:
                break

            if values['state'] == 'confirmed':
                if values['healthprof']:
                    healthprof = Healthprof(values['healthprof'])
                    if (healthprof.party.internal_user and
                            healthprof.party.internal_user.calendar):
                        patient = Patient(values['patient'])
                        appointment_date_end = None
                        if values.get('appointment_date_end'):
                            appointment_date_end = \
                                values['appointment_date_end']
                        events = Event.create([{
                            'dtstart': values['appointment_date'],
                            'dtend': appointment_date_end,
                            'calendar':
                                healthprof.party.internal_user.calendar.id,
                            'summary': patient.party.rec_name,
                            'description': values['comments'],
                        }])
                        values['event'] = events[0].id
        return super(Appointment, cls).create(vlist)

    @classmethod
    def write(cls, appointments, values):
        pool = Pool()
        Event = pool.get('calendar.event')
        Patient = pool.get('gnuhealth.patient')
        Healthprof = pool.get('gnuhealth.healthprofessional')

        for appointment in appointments:
            # Update caldav event
            if appointment.event and ('healthprof' not in values):
                if 'appointment_date' in values:
                    Event.write([appointment.event], {
                        'dtstart': values['appointment_date'],
                    })
                if 'appointment_date_end' in values:
                    Event.write([appointment.event], {
                        'dtend': values['appointment_date_end'],
                    })
                if 'patient' in values:
                    patient = Patient(values['patient'])
                    Event.write([appointment.event], {
                        'summary': patient.party.rec_name,
                    })
                if 'comments' in values:
                    Event.write([appointment.event], {
                        'description': values['comments'],
                    })

            else:
                # Move the event to the new health professional
                if appointment.event and ('healthprof' in values):
                    current_event = [appointment.event]
                    if appointment.healthprof.party.internal_user:
                        healthprof = Healthprof(values['healthprof'])
                        if healthprof.party.internal_user.calendar:
                            # Health professional has calendar
                            patient = appointment.patient.party.rec_name
                            comments = ''
                            if 'comments' in values:
                                comments = values['comments']
                            else:
                                comments = appointment.comments
                            if 'appointment_date' in values:
                                appointment_date = values['appointment_date']
                            else:
                                appointment_date = appointment.appointment_date
                            if 'appointment_date_end' in values:
                                appointment_date_end = \
                                    values['appointment_date_end']
                            else:
                                appointment_date_end = \
                                    appointment.appointment_date_end
                            events = Event.create([{
                                'dtstart': appointment_date,
                                'dtend': appointment_date_end,
                                'calendar':
                                    healthprof.party.internal_user.calendar.id,
                                'summary':
                                    patient,
                                'description': comments,
                            }])
                            values['event'] = events[0].id

                    # Delete the event from the current health professional
                    # after it has been transfer to the new healthpfof
                    Event.delete(current_event)
        return super(Appointment, cls).write(appointments, values)

    @classmethod
    def delete(cls, appointments):
        Event = Pool().get('calendar.event')

        for appointment in appointments:
            if appointment.event:
                Event.delete([appointment.event])
        return super(Appointment, cls).delete(appointments)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                      HEALTH FEDERATION package                        #
#                  health_federation.py: main module                    #
#########################################################################

from trytond.model import ModelView, ModelSQL, ModelSingleton, fields, Unique
from trytond.pool import Pool
from trytond.pyson import Eval, Equal
import requests
import json
from uuid import uuid4
from trytond.rpc import RPC
from datetime import datetime, date
from trytond.modules.health.core import get_institution
from trytond.i18n import gettext

from .exceptions import (
    NeedLoginCredentials, ServerAuthenticationError,
    ThalamusConnectionError, ThalamusConnectionOK,
    NoInstitution)

__all__ = [
    'FederationNodeConfig', 'FederationQueue', 'FederationObject',
    'PartyFed', 'PoLFed']


def set_fsync(model, records, flag):
    # Sets or disables the fsync flag that enables the record to be
    # enter in the Federation queue
    vals = {'fsync': flag}
    model.write(records, vals)


class FederationNodeConfig(ModelSingleton, ModelSQL, ModelView):
    'Federation Node Configuration'
    __name__ = 'gnuhealth.federation.config'

    host = fields.Char(
        'Thalamus server', required=True,
        help="GNU Health Thalamus server")
    port = fields.Integer('Port', required=True, help="Thalamus port")
    user = fields.Char(
        'User', required=True,
        help="Admin associated to this institution")
    password = fields.Char(
        'Password', required=True,
        help="Password of the institution admin user in the Federation")
    ssl = fields.Boolean('SSL', help="Use encrypted communication via SSL")
    verify_ssl = fields.Boolean(
        'Verify SSL cert',
        help="Check this option if your certificate has been emitted"
             " by a CA authority. If it is a self-signed certifiate"
             " leave it unchecked")
    enabled = fields.Boolean(
        'Enabled', help="Mark if the node is active"
        " in the Federation")

    # TODO: Check the associated institution and use as
    # a default value its id

    @staticmethod
    def default_host():
        return 'localhost'

    @staticmethod
    def default_port():
        return 8443

    @staticmethod
    def default_ssl():
        return True

    @staticmethod
    def default_enabled():
        return True

    @staticmethod
    def default_verify_ssl():
        return False

    @staticmethod
    def default_database():
        return 'federation'

    @classmethod
    def __setup__(cls):
        super(FederationNodeConfig, cls).__setup__()

        cls._buttons.update({
            'test_connection': {}
        }),

    @classmethod
    @ModelView.button
    def test_connection(cls, argvs):
        """ Make the connection test to Thalamus Server
            from the GNU Health HMIS using the institution
            associated admin and the related credentials
        """

        host, port, user, password, ssl_conn, verify_ssl = \
            argvs[0].host, argvs[0].port,  \
            argvs[0].user, argvs[0].password, argvs[0].ssl, \
            argvs[0].verify_ssl

        if (ssl_conn):
            protocol = 'https://'
        else:
            protocol = 'http://'

        if (not user or not password):
            raise NeedLoginCredentials(
                gettext('health_federation.msg_need_login_credentials')
            )

        url = protocol + host + ':' + str(port) + '/people/' + user

        try:
            conn = requests.get(
                url,
                auth=(user, password), verify=verify_ssl)

        except BaseException:
            raise ServerAuthenticationError(
                gettext('health_federation.msg_server_authentication_error')
            )

        if conn:
            raise ThalamusConnectionOK(
                gettext('health_federation.msg_thalamus_connection_ok')
            )

        else:
            raise ThalamusConnectionError(
                gettext('health_federation.msg_thalamus_connection_error')
            )

    @classmethod
    def get_conn_params(cls):
        # Retrieve the connection information for Thalamus
        # Retrieve the information from the Singleton object

        TInfo = Pool().get(cls.__name__)(1)
        host = TInfo.host
        port = TInfo.port
        user = TInfo.user
        password = TInfo.password
        ssl_conn = TInfo.ssl
        verify_ssl = TInfo.verify_ssl

        if (ssl_conn):
            protocol = 'https://'
        else:
            protocol = 'http://'

        return host, port, user, password, ssl_conn, verify_ssl, protocol


class FederationQueue(ModelSQL, ModelView):
    'Federation Queue'
    __name__ = 'gnuhealth.federation.queue'

    msgid = fields.Char(
        'Message ID', required=True,
        help="Message UID")
    model = fields.Char('Model', required=True, help="Source Model")
    node = fields.Char(
        'Node', required=True,
        help="The originating node id")

    time_stamp = fields.Char(
        'Timestamp', required=True,
        help="UTC timestamp at the moment of writing record on the node")

    federation_locator = fields.Char(
        'Fed ID',
        help="Unique locator in Federation, "
        "such as person Federation account or Page of Life code")

    args = fields.Text(
        'Arguments', required=True,
        help="Arguments")

    args_preview = fields.Function(
        fields.Text('Preview'),
        "get_args_preview_string")

    def get_args_preview_string(self, name):
        try:
            data = json.loads(self.args)
            return json.dumps(data, indent=4, ensure_ascii=False)
        except BaseException:
            return "Fail to parse json string of args field."

    method = fields.Selection([
        (None, ''),
        ('POST', 'POST'),
        ('PATCH', 'PATCH'),
        ('DELETE', 'DELETE'),
        ('GET', 'GET'),
    ], 'Method', required=True, sort=False)

    state = fields.Selection([
        (None, ''),
        ('queued', 'Queued'),
        ('sent', 'Sent'),
        ('failed', 'Failed'),
    ], 'Status', sort=False)

    url_suffix = fields.Char(
        'URL suffix',
        help="suffix to be passed to the URL")

    @staticmethod
    def default_node():
        # Get the Institution code as the originating node.
        HealthInst = Pool().get('gnuhealth.institution')
        institution = get_institution()

        if (institution):
            # Get the institution code associated to the ID
            institution_code = HealthInst(institution).code

        else:
            raise NoInstitution(
                gettext('health_federation.msg_no_institution')
            )

        return institution_code

    @classmethod
    def send_record(cls, record):
        rc = False

        host, port, user, password, ssl_conn, verify_ssl, protocol = \
            FederationNodeConfig.get_conn_params()

        if (record.method == 'PATCH'):
            if (record.federation_locator):
                # Traverse each resource and its data fields.
                # arg : Dictionary for each of the data elements in the
                #       list of values
                # [{resource, fields[{name, value}]

                for arg in json.loads(record.args):
                    vals = {}
                    modification_info = {}

                    # Include the modification information
                    modification_info = {
                        'user': user,
                        'timestamp': record.time_stamp,
                        'node': record.node}

                    vals['modification_info'] = modification_info

                    url = protocol + host + ':' + str(port)

                    resource, fields = arg['resource'], \
                        arg['fields']

                    # Add resource and instance to URL
                    url = url + '/' + resource + '/' + record.url_suffix

                    for field in fields:
                        fname, fvalue = field['name'], field['value']
                        vals[fname] = fvalue

                    send_data = requests.request(
                        'PATCH', url,
                        data=json.dumps(vals),
                        auth=(user, password), verify=verify_ssl)

                    if (send_data):
                        rc = True
            else:
                print("No federation record locator found .. no update")

        # Create the record on the Federation
        if (record.method == 'POST'):
            if (record.federation_locator):
                # Traverse each resource and its data fields.
                # arg : Dictionary for each of the data elements in the
                #       list of values
                # [{resource, fields[{name, value}]

                for arg in json.loads(record.args):
                    vals = {}
                    creation_info = {}

                    # Include the creation information
                    creation_info = {
                        'user': user,
                        'timestamp': record.time_stamp,
                        'node': record.node}

                    vals['creation_info'] = creation_info

                    url = protocol + host + ':' + str(port)

                    resource, fields = arg['resource'], \
                        arg['fields']

                    # Add resource and instance to URL
                    url = url + '/' + resource + '/' + record.url_suffix

                    for field in fields:
                        fname, fvalue = field['name'], field['value']
                        vals[fname] = fvalue

                    send_data = requests.request(
                        'POST', url,
                        data=json.dumps(vals),
                        auth=(user, password), verify=verify_ssl)

                    if (send_data):
                        rc = True
            else:
                print("No federation record locator found .. no update")

        return rc

    @classmethod
    def parse_fields(cls, values, action, fields):
        ''' Returns, depending on the action, the fields that will be
            passed as arguments to Thalamus
        '''
        # Fedvals : Values to send to the federation, that contain the
        # resource, the field, and the value of such field.

        field_mapping = fields.split(',')
        resources = []
        fedvals = []

        fed_key = {}
        n = 0
        for val in field_mapping:
            # Retrieve the field name, the federation resource name
            # and the associated federation resource field .
            # string of the form field:fed_resource:fed_field
            field, fed_resource, fed_field = val.split(':')
            if (field in values.keys()):
                # Check that the local field is on shared federation list
                # and retrieve the equivalent federation field name

                # Optimize the process, by not duplicating the resources
                # Each resource will appear just once, with the list of
                # its fields.
                # Create a new item in the resource list
                # to be returned, if this is not yet there.

                # Serialize the date, datetime objects to be JSON compat
                if isinstance(values[field], (datetime, date)):
                    values[field] = values[field].isoformat()

                if fed_resource not in resources:
                    fed_key = {
                        "resource": fed_resource,
                        "fields": [
                            {
                                "name": fed_field,
                                "value": values[field]
                            }]
                    }

                    resources.append(fed_resource)
                    fedvals.append(fed_key)

                else:
                    # Otherwise, if the resource exists on the list,
                    # we add the new field values to it.

                    n = 0
                    for record in fedvals:
                        if fed_resource == record['resource']:
                            fedvals[n]['fields'].append(
                                {"name": fed_field,
                                 "value": values[field]})
                        n = n + 1

        return fedvals

    @classmethod
    def enqueue(cls, model, federation_loc, time_stamp, node, values,
                action, url_suffix):

        fields = FederationObject.get_object_fields(model)
        # Federation locator : Unique ID of the resource
        # such as personal federation account or institution ID
        # it depends on the resource (people, PoL, institution, ... )

        # Remove spaces and newlines from fields
        fields = fields.replace(" ", "").replace("\n", "")

        if (fields):
            # retrieve the federation field names and values in a dict
            fields_to_enqueue = cls.parse_fields(values, action, fields)
            # Continue the enqueue process with the fields

            if fields_to_enqueue:
                rec = []
                vals = {}
                vals['msgid'] = str(uuid4())
                vals['model'] = model
                vals['time_stamp'] = str(time_stamp)
                vals['args'] = json.dumps(fields_to_enqueue)
                vals['method'] = action
                vals['state'] = 'queued'
                vals['federation_locator'] = federation_loc
                vals['url_suffix'] = url_suffix
                rec.append(vals)
                # Write the record to the enqueue list
                cls.create(rec)

    @classmethod
    def __setup__(cls):
        super(FederationQueue, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('msgid_uniq', Unique(t, t.msgid),
                'The Message ID must be unique !'), ]

        cls._buttons.update({
            'send': {'invisible': Equal(Eval('state'), 'sent')}
        })

    @classmethod
    @ModelView.button
    def send(cls, records):
        # Verify and send each record to Thalamus individually
        # It allows to get specific status on each operation
        # record: individual record on gnuhealth.federation.queue model
        for record in records:
            rec = []
            res = cls.send_record(record)
            if res:
                rec.append(record)
                cls.write(rec, {'state': 'sent'})
            else:
                cls.write(rec, {'state': 'failed'})


class FederationObject(ModelSQL, ModelView):
    'Federation Object'
    __name__ = 'gnuhealth.federation.object'

    model = fields.Char('Model', help="Local Model")

    enabled = fields.Boolean(
        'Enabled', help="Check if the model"
        " is active to participate on the Federation")

    fields = fields.Text(
        'Fields', help="Contains a list of "
        "the local model fields that participate on the federation.\n"
        "Each line will have the format field:endpoint:key")

    @staticmethod
    def default_enabled():
        return True

    @classmethod
    def get_object_fields(cls, obj):
        model, = cls.search_read(
            [("model", "=", obj)],
            limit=1, fields_names=['fields', 'enabled'])

        # If the model exist on the Federation Object,
        # and is currently enabled, return the field names and their
        # equivalents on the Federation
        if (model and model['enabled']):
            return model['fields']

    @classmethod
    def __setup__(cls):
        super(FederationObject, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('model_uniq', Unique(t, t.model),
             'The Model is already defined !')
        ]
        cls.__rpc__.update({
            'get_object_fields': RPC(check_access=False),
        })


class PartyFed(ModelSQL):
    """
    Demographics Model to participate on the Federation

    Methods : write (PATCH), create (POST)
    """

    __name__ = 'party.party'

    # Controls if the updated information will be sent to the Federation.
    #   True: Info will be sent to the federation queue
    #   False: The info is up-todate with the federation or is local
    fsync = fields.Boolean(
        'Fsync',
        help="If active, this record information will"
             " be sent to the Federation")

    @staticmethod
    def default_fsync():
        return True

    @classmethod
    def write(cls, parties, values):
        # First exec the Party class write method from health package
        super(PartyFed, cls).write(parties, values)

        for party in parties:
            action = "PATCH"
            # Retrieve federation account (for people only)
            fed_acct = party.federation_account
            fsync = party.fsync
            # Verify that the person has a Federation account.
            # and the fsync flag is set for the record
            if (fed_acct and values):
                if (fsync or 'fsync' not in values.keys()):
                    # Start Enqueue process
                    node = None
                    # Because du_address is a functional field
                    # does not exist at DB level, we always pass the
                    # latest / current value
                    values['du_address'] = party.du_address
                    time_stamp = party.write_date
                    url_suffix = fed_acct
                    FederationQueue.enqueue(
                        cls.__name__,
                        fed_acct, time_stamp, node, values, action,
                        url_suffix)

                    # Unset the fsync flag locally once the info has been
                    # sent to the Federation queue
                    set_fsync(cls, parties, False)

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]

        # Execute first the creation of party
        parties = super(PartyFed, cls).create(vlist)

        for values in vlist:
            fed_acct = parties[0].federation_account

            # Check if the record has just been imported of modified
            # from the federation. In that case, skip sending it
            # to the federation queue

            fsync = parties[0].fsync
            # If the user has a federation ID, then enqueue the
            # record to be sent and created in the Federation

            if fed_acct and fsync:
                action = "POST"
                node = None

                # Because du_address is a functional field
                # does not exist at DB level, we always pass the
                # latest / current value
                values['du_address'] = parties[0].du_address

                url_suffix = fed_acct
                time_stamp = parties[0].create_date
                FederationQueue.enqueue(
                    cls.__name__,
                    fed_acct, time_stamp, node, values,
                    action, url_suffix)

                # Unset the fsync flag locally once the info has been
                # sent to the Federation queue
                if (fsync):
                    set_fsync(cls, parties, False)

        return parties


class PoLFed(ModelSQL):
    """
    Page of Life Model to participate on the Federation

    Methods : write (PATCH), create (POST)
    """

    __name__ = 'gnuhealth.pol'

    # Controls if the updated information will be sent to the Federation.
    #   True: Info will be sent to the federation queue
    #   False: The info is up-todate with the federation or is local
    fsync = fields.Boolean(
        'Fsync',
        help=" If active, this record information will"
             " be sent to the Federation")

    @staticmethod
    def default_fsync():
        return True

    @classmethod
    def write(cls, pols, values):
        super(PoLFed, cls).write(pols, values)

        for pol in pols:
            action = "PATCH"
            # Retrieve page of life unique ID into fed_identifier
            fed_identifier = pol.page
            federation_account = pol.federation_account

            fsync = pol.fsync
            # Verify that the page has the federation identifier.
            # and the fsync flag is set for the record
            if (fed_identifier and values):
                if (fsync or 'fsync' not in values.keys()):
                    # Start Enqueue process
                    node = None
                    time_stamp = pol.write_date
                    url_suffix = federation_account + '/' + fed_identifier
                    FederationQueue.enqueue(
                        cls.__name__,
                        fed_identifier, time_stamp, node, values,
                        action, url_suffix)

                    # Unset the fsync flag locally once the info has been
                    # sent to the Federation queue
                    set_fsync(cls, pols, False)

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]

        # Execute first the creation of PoL
        pols = super(PoLFed, cls).create(vlist)

        for values in vlist:
            fed_identifier = pols[0].page
            federation_account = pols[0].federation_account
            # Check if the record has just been imported of modified
            # from the federation. In that case, skip sending it
            # to the federation queue

            fsync = pols[0].fsync
            # If the page has a federation ID, then enqueue the
            # record to be sent and created in the Federation

            if fed_identifier and fsync:
                action = "POST"
                node = None
                time_stamp = pols[0].create_date
                url_suffix = federation_account + '/' + fed_identifier
                FederationQueue.enqueue(
                    cls.__name__,
                    fed_identifier, time_stamp, node, values,
                    action, url_suffix)

                # Unset the fsync flag locally once the info has been
                # sent to the Federation queue
                if (fsync):
                    set_fsync(cls, pols, False)

        return pols



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                 HEALTH PEDIATRICS_GROWTH_CHARTS_WHO package           #
#                  __init__.py: Package declaration file                #
#########################################################################
from trytond.pool import Pool
from . import health_pediatrics_growth_charts_who
from . import wizard
from . import report


def register():
    Pool.register(
        health_pediatrics_growth_charts_who.PediatricsGrowthChartsWHO,
        wizard.wizard_health_pediatrics_growth_charts_who.
        OpenPediatricsGrowthChartsWHOReportStart,
        module='health_pediatrics_growth_charts_who', type_='model')
    Pool.register(
        wizard.wizard_health_pediatrics_growth_charts_who.
        OpenPediatricsGrowthChartsWHOReport,
        module='health_pediatrics_growth_charts_who', type_='wizard')
    Pool.register(
        report.report_health_pediatrics_growth_charts_who.
        PediatricsGrowthChartsWHOReport,
        report.report_health_pediatrics_growth_charts_who.
        WeightForAge,
        report.report_health_pediatrics_growth_charts_who.
        LengthHeightForAge,
        report.report_health_pediatrics_growth_charts_who.
        BMIForAge,
        module='health_pediatrics_growth_charts_who', type_='report')
